!	A simple optimizer for T -- Copyright (C) 1997,1998 Nils M. Holm
!	<fs29@rumms.uni-mannheim.de>
!
!	This is a simple optimizer for the symbolic frontend of
!	the T compiler. It reads symbolic input from stdin and
!	writes (optimized) symbolic output to stdout.
!	TSOPT builds a tree structure from its input which is
!	suitable for a wide range of different optimizations.
!	The framework provided in this file may be used to
!	implement expression statement and basic block level
!	optimizations. Since the generated tree is destroyed
!	after processing a routine, no global optimizations may
!	be done.
!	Currently, TSOPT does the following transformations:
!	(N = constant, X = variable)
!
!	All constant expressions are evaluated and replaced with
!	numeric results.
!
!	The following redundant operations are eliminated:
!		X*0, 0*X	->	0
!		X*1, 1*X, X/1	->	X
!		X/0		->	ERROR
!		0/X		->	0
!		X mod 1		->	0
!		X mod 0		->	ERROR
!		0 mod X		->	0
!		X+0, 0+X	->	X
!		X-0		->	X
!		X&0, 0&X	->	0
!		X&%1, %1&X	->	X
!		X|0, 0|X	->	X
!		X|%1, %1|X	->	%1
!		0<<X, 0>>X	->	0
!		X>>0, X<<0	->	X
!		X[0]		->	X (dereferenced)
!
!	The following operations are reduced:
!		0-X		->	-X
!		X^%1, %1^X	->	~X
!		0=X, X=0	->	\X
!		X*N, N*X	->	X<<LOG2(N), if LOG2(N) e N
!		X/N		->	X>>LOG2(N), if LOG2(N) e N
!		X := X+N,
!		X := N+X	->	inc(X,N)
!		X := X-N	->	inc(X,-N)
!
!	The last three transformations are possible only because
!	both, the shift operations and the term operations, are
!	always unsigned in T.

const	BPW = 2;

const	BUFLEN		= 1025;
const	NODESPACE	= 16000;

const	ILABEL='L', IPOP='x', IDECL='d', ILDG='g', ILDGV='V', ILDL='p',
	ILDLV='v', IIND='i', IDEREF='X', ISAVG='G', ISAVL='P', ISTORE='S',
	ISTACK='s', IUNSTK='u', ICALL='c', INUM='n', ISTR='$', ILDLAB='l',
	IDATA='D', INEG='_', ILNOT='\\', IBNOT='~', IMUL='*', IDIV='/',
	IMOD='m', IADD='+', ISUB='-', IBAND='&', IBOR='|', IBXOR='^',
	IBSHL='{', IBSHR='}', IEQU='=', INEQU='#', ILESS='<', IGRTR='>',
	ILTEQ='(', IGTEQ=')', IINCG='Z', IINCL='z', IBRF='b', IBRT='B',
	IJUMP='j', IUFOR='f', IDFOR='F', IHDR='r', IEND='R', IHALT='h',
	IINIT='T', IPUSH='!', INOP='.';
const	IGLUE='0';

const	EOF=%1, NUMBER=0;

const	NTOKEN	= 0,
	NVALUE	= 1,
	NVAL2	= 2,
	NNVAL	= 3,
	NLEFT	= 4,
	NRIGHT	= 5,
	NODELEN	= 6;

var	Buffer[BUFLEN];
var	Bp, Ep, Pos;
var	Obuf[BUFLEN];
var	Obp;
var	Nodes[NODESPACE], Nt;
var	Token;
var	Value;
var	Maxnodes;


scan() do
	var	c, state, sign;

	state := 0;
	Value := 0;
	while (1) do
		if (Bp >= Ep) do
			Ep := reads(Buffer, BUFLEN-1);
			Bp := 0;
			if (\Ep) do
				Token := EOF;
				return EOF;
			end
		end
		c := Buffer[Bp];
		Bp := Bp+1;
		Pos := Pos+1;
		ie (\state) do
			ie ('0' <= c /\ c <= '9' \/ c = '%') do
				Value := c='%'-> 0: c-'0';
				state := 1;
				sign := c = '%';
			end
			else do
				Token := c;
				return c;
			end
		end
		else do
			ie ('0' <= c /\ c <= '9') do
				Value := Value * 10 + (c-'0');
			end
			else do
				Bp := Bp-1;
				Pos := Pos-1;
				Token := NUMBER;
				if (sign) Value := -Value;
				return NUMBER;
			end
		end
	end
end


length(s) do
	var	i;

	i := 0;
	while (s[i]) i := i+1;
	return i;
end


copy(s, d) do
	var	i;

	i := 0;
	while (s[i]) do
		d[i] := s[i];
		i := i+1;
	end
	d[i] := 0;
end


report(msg) do
	var	ch[2];

	select(1, 2);
	writes("tsopt: "); writes(msg);
	writes(" [P=");
	writes(ntoa(Pos, 0));
	writes(", T=");
	ch[0] := Token;
	ch[1] := 0;
	ie (Token = NUMBER)
		writes("NUMBER");
	else ie (Token = EOF)
		writes("EOF");
	else
		writes(ch);
	writes(", V=");
	writes(ntoa(Value, 0));
	writes("]");
	newline();
	halt;
end


newnode(t, v, l, r) do
	var	n;

	if (Nt + NODELEN >= NODESPACE) report("out of nodes");
	n := @Nodes[Nt];
	Nt := Nt+NODELEN;
	n[NTOKEN] := t;
	n[NVALUE] := v;
	n[NVAL2] := 0;
	n[NNVAL] := 0;
	n[NLEFT] := l;
	n[NRIGHT] := r;
	return n;
end


num() do
	scan();
	if (Token \= NUMBER) report("numeric parameter expected");
	return Value;
end


label() do
	scan();
	if (Token \= ILABEL) report("label expected");
	num();
end


semi() do
	scan();
	if (Token \= ';') report("separator expected");
end


bwrite(s) do
	var	k;

	if (\s) do
		writes(Obuf); Obp := 0;
		return 0;
	end
	k := length(s);
	if (Obp + k >= BUFLEN) do
		writes(Obuf); Obp := 0;
	end
	copy(s, @Obuf[Obp]);
	Obp := Obp + k;
end


ztoa(n) do
	var	z;

	z := ntoa(n, 0);
	if (z[0] = '-') z[0] := '%';
	return z;
end


emitn(n) do
	bwrite(ztoa(n));
end


emit(i) do
	var	ch[2];

	ch[0] := i;
	ch[1] := 0;
	bwrite(ch);
end


mark(l) do
	emit(ILABEL);
	emitn(l);
end


string() do
	var	i, k;

	emit('$');
	k := num();
	emitn(k);
	scan();
	if (Token \= '$') report("string data expected");
	emit('$');
	for (i=0, k) do
		emitn(num());
		if (i<k-1) do
			semi();
			emit(';');
		end
	end
end


cleanup() do
	if (Nt > Maxnodes) Maxnodes := Nt;
	Nt := 0;
end


writenode(n) do
	var	ch[2];
	var	l;

	if (n[NTOKEN] = IGLUE) return 0;
	ch[0] := n[NTOKEN];
	ch[1] := 0;
	bwrite(ch);
	l := n[NNVAL];
	if (l & 256) bwrite("L");
	l := l & 255;
	if (l > 0) bwrite(ztoa(n[NVALUE]));
	if (l > 1) do
		bwrite(";");
		bwrite(ztoa(n[NVAL2]));
	end
end


dumptree(t, l) do
	var	i;

	if (\t) return 0;
	for (i=0, l) bwrite(".");
	ie (t[NTOKEN] = IGLUE)
		bwrite("0");
	else
		writenode(t);
	bwrite("\n");
	if (t[NLEFT]) bwrite("L->");
	dumptree(t[NLEFT], l+1);
	if (t[NRIGHT]) bwrite("R->");
	dumptree(t[NRIGHT], l+1);
end


emittree(t) do
	if (\t) return 0;
	emittree(t[NLEFT]);
	emittree(t[NRIGHT]);
	writenode(t);
end


fold_unary(t, op) do
	t[NTOKEN] := INUM;
	t[NNVAL] := 1;
	ie (op = INEG) t[NVALUE] := -t[NLEFT][NVALUE];
	else ie (op = ILNOT) t[NVALUE] := \t[NLEFT][NVALUE];
	else t[NVALUE] := ~t[NLEFT][NVALUE];
	t[NLEFT] := 0;
	return t;
end


fold_binary(t, op) do
	t[NTOKEN] := INUM;
	t[NNVAL] := 1;
	ie (op = IMUL) t[NVALUE] :=
		t[NLEFT][NVALUE] * t[NRIGHT][NVALUE];
	else ie (op = IDIV) do
			if (\t[NRIGHT][NVALUE])
				report("constant divide by zero");
			t[NVALUE] := t[NLEFT][NVALUE] / t[NRIGHT][NVALUE];
		end
	else ie (op = IMOD) do
			if (\t[NRIGHT][NVALUE])
				report("constant divide by zero");
		t[NVALUE] := t[NLEFT][NVALUE] mod t[NRIGHT][NVALUE];
	end
	else ie (op = IADD) t[NVALUE] :=
		t[NLEFT][NVALUE] + t[NRIGHT][NVALUE];
	else ie (op = ISUB) t[NVALUE] :=
		t[NLEFT][NVALUE] - t[NRIGHT][NVALUE];
	else ie (op = IBAND) t[NVALUE] :=
		t[NLEFT][NVALUE] & t[NRIGHT][NVALUE];
	else ie (op = IBOR) t[NVALUE] :=
		t[NLEFT][NVALUE] | t[NRIGHT][NVALUE];
	else ie (op = IBXOR) t[NVALUE] :=
		t[NLEFT][NVALUE] ^ t[NRIGHT][NVALUE];
	else ie (op = IBSHL) t[NVALUE] :=
		t[NLEFT][NVALUE] << t[NRIGHT][NVALUE];
	else ie (op = IBSHR) t[NVALUE] :=
		t[NLEFT][NVALUE] >> t[NRIGHT][NVALUE];
	else ie (op = IEQU) t[NVALUE] :=
		t[NLEFT][NVALUE] = t[NRIGHT][NVALUE];
	else ie (op = INEQU) t[NVALUE] :=
		t[NLEFT][NVALUE] \= t[NRIGHT][NVALUE];
	else ie (op = ILESS) t[NVALUE] :=
		t[NLEFT][NVALUE] < t[NRIGHT][NVALUE];
	else ie (op = IGRTR) t[NVALUE] :=
		t[NLEFT][NVALUE] > t[NRIGHT][NVALUE];
	else ie (op = ILTEQ) t[NVALUE] :=
		t[NLEFT][NVALUE] <= t[NRIGHT][NVALUE];
	else
		t[NVALUE] :=
		t[NLEFT][NVALUE] >= t[NRIGHT][NVALUE];
	t[NLEFT] := 0;
	t[NRIGHT] := 0;
	return t;
end


power_of_2(v) do
	var	x, r;

	x := 1;
	r := 2;
	while (r) do
		if (v = r) return x;
		x := x+1;
		r := r*2;
	end;
	return 0;
end


fold_left(t, op) do
	const	TVAL = 1, TEXP = 2;
	var	v, x, tr;

	tr := 0;
	v := t[NLEFT][NVALUE];
	ie (v = 0) do
		ie (	op = IMUL \/ op = IBAND \/ op = IBSHL \/ op = IBSHR \/
			op = IDIV \/ op = IMOD )
		do
			v := 0; tr := TVAL;
		end
		else ie (op = IADD \/ op = IBOR) do
			tr := TEXP;
		end
		else ie (op = ISUB) do
			t[NTOKEN] := INEG;
			t[NLEFT] := t[NRIGHT];
			t[NRIGHT] := 0;
		end
		else if (op = IEQU) do
			t[NTOKEN] := ILNOT;
			t[NLEFT] := t[NRIGHT];
			t[NRIGHT] := 0;
		end
	end
	else ie (v = 1) do
		if (op = IMUL) do
			tr := TEXP;
		end
	end
	else ie (v = %1) do
		ie (op = IBAND) do
			tr := TEXP;
		end
		else ie (op = IBOR) do
			v := %1; tr := TVAL;
		end
		else if (op = IBXOR) do
			t[NTOKEN] := ILNOT;
			t[NLEFT] := t[NRIGHT];
			t[NRIGHT] := 0;
		end
	end
	else do
! DISABLED DUE TO CHANGE TO SIGNED DIVIDE AND MODULO
!		v := power_of_2(v);
!		if (v) do
!			if (op \= IMUL) return t;
!			t[NTOKEN] := IBSHL;
!			x := t[NLEFT];
!			t[NLEFT] := t[NRIGHT];
!			t[NRIGHT] := x;
!			x[NVALUE] := v;
!		end
	end
	ie (tr = TVAL) do
		t[NTOKEN] := INUM;
		t[NNVAL] := 1;
		t[NVALUE] := v;
		t[NLEFT] := 0;
		t[NRIGHT] := 0;
	end
	else if (tr = TEXP) do
		t[NTOKEN] := t[NRIGHT][NTOKEN];
		t[NVALUE] := t[NRIGHT][NVALUE];
		t[NNVAL] := t[NRIGHT][NNVAL];
		t[NVAL2] := t[NRIGHT][NVAL2];
		t[NLEFT] := t[NRIGHT][NLEFT];
		t[NRIGHT] := t[NRIGHT][NRIGHT];
	end
	return t;
end


fold_right(t, op) do
	const	TVAL = 1, TEXP = 2;
	var	v, tr;

	tr := 0;
	v := t[NRIGHT][NVALUE];
	ie (v = 0) do
		ie (op = IMUL \/ op = IBAND) do
			v := 0; tr := TVAL;
		end
		else ie (op = IBSHL \/ op = IBSHR) do
			tr := TEXP;
		end
		else ie (op = IDIV \/ op = IMOD ) do
			report("constant divide by zero");
		end
		else ie (op = IADD \/ op = IBOR \/ op = ISUB) do
			tr := TEXP;
		end
		else if (op = IEQU) do
			t[NTOKEN] := ILNOT;
			t[NRIGHT] := 0;
		end
	end
	else ie (v = 1) do
		ie (op = IMUL \/ op = IDIV) do
			tr := TEXP;
		end
		else if (op = IMOD) do
			v := 0; tr := TVAL;
		end
	end
	else ie (v = %1) do
		ie (op = IBAND) do
			tr := TEXP;
		end
		else ie (op = IBOR) do
			v := %1; tr := TVAL;
		end
		else if (op = IBXOR) do
			t[NTOKEN] := ILNOT;
			t[NRIGHT] := 0;
		end
	end
	else do
! DISABLED DUE TO CHANGE TO SIGNED DIVIDE AND MODULO
!		v := power_of_2(v);
!		if (v) do
!			ie (op = IMUL)
!				t[NTOKEN] := IBSHL;
!			else ie (op = IDIV)
!				t[NTOKEN] := IBSHR;
!			else
!				return t;
!			t[NRIGHT][NVALUE] := v;
!		end
	end
	ie (tr = TVAL) do
		t[NTOKEN] := INUM;
		t[NNVAL] := 1;
		t[NVALUE] := v;
		t[NLEFT] := 0;
		t[NRIGHT] := 0;
	end
	else if (tr = TEXP) do
		t[NTOKEN] := t[NLEFT][NTOKEN];
		t[NVALUE] := t[NLEFT][NVALUE];
		t[NNVAL] := t[NLEFT][NNVAL];
		t[NVAL2] := t[NLEFT][NVAL2];
		t[NRIGHT] := t[NLEFT][NRIGHT];
		t[NLEFT] := t[NLEFT][NLEFT];
	end
	return t;
end


reduce_inc(t) do
	var	l, r, b, sign;

	b := t[NRIGHT];
	l := t[NLEFT][NRIGHT][NRIGHT][NLEFT];
	r := t[NLEFT][NRIGHT][NRIGHT][NRIGHT];
	if (	l[NLEFT] \/ l[NRIGHT] \/ r[NLEFT] \/ r[NRIGHT] \/
		b[NLEFT] \/ b[NRIGHT]
	)
		return t;
	sign := t[NLEFT][NRIGHT][NRIGHT][NTOKEN] = ISUB;
	ie (	l[NTOKEN] = ILDG /\ b[NTOKEN] = ISAVG /\
		l[NVALUE] = b[NVALUE] /\
		r[NTOKEN] = INUM )
	do
		t[NRIGHT][NTOKEN] := IINCG;
		t[NRIGHT][NNVAL] := 2;
		t[NRIGHT][NVAL2] := sign-> -r[NVALUE]: r[NVALUE];
		t[NLEFT][NRIGHT] := 0;
	end
	else ie (l[NTOKEN] = ILDL /\ b[NTOKEN] = ISAVL /\
		l[NVALUE] = b[NVALUE] /\
		r[NTOKEN] = INUM )
	do
		t[NRIGHT][NTOKEN] := IINCL;
		t[NRIGHT][NNVAL] := 2;
		t[NRIGHT][NVAL2] := sign-> -r[NVALUE]: r[NVALUE];
		t[NLEFT][NRIGHT] := 0;
	end
	else ie (r[NTOKEN] = ILDG /\ b[NTOKEN] = ISAVG /\
		r[NVALUE] = b[NVALUE] /\
		l[NTOKEN] = INUM /\ \sign)
	do
		t[NRIGHT][NTOKEN] := IINCG;
		t[NRIGHT][NNVAL] := 2;
		t[NRIGHT][NVAL2] := sign-> -l[NVALUE]: l[NVALUE];
		t[NLEFT][NRIGHT] := 0;
	end
	else if (r[NTOKEN] = ILDL /\ b[NTOKEN] = ISAVL /\
		r[NVALUE] = b[NVALUE] /\
		l[NTOKEN] = INUM /\ \sign)
	do
		t[NRIGHT][NTOKEN] := IINCL;
		t[NRIGHT][NNVAL] := 2;
		t[NRIGHT][NVAL2] := sign-> -l[NVALUE]: l[NVALUE];
		t[NLEFT][NRIGHT] := 0;
	end
	return t;
end


optimize(t) do
	var	op, v;

	if (\t) return t;
	t[NLEFT] := optimize(t[NLEFT]);
	t[NRIGHT] := optimize(t[NRIGHT]);
	op := t[NTOKEN];
	ie (	op = INEG \/ op = ILNOT \/ op = IBNOT )
	do
		if (t[NLEFT][NTOKEN] = INUM) t := fold_unary(t, op);
	end
	else ie (op = IMUL \/ op = IDIV \/ op = IMOD \/ op = IADD \/
		op = ISUB \/ op = IBAND \/ op = IBOR \/ op = IBXOR \/
		op = IBSHL \/ op = IBSHR \/ op = IEQU \/ op = INEQU \/
		op = ILESS \/ op = IGRTR \/ op = ILTEQ \/ op = IGTEQ )
	do
		ie (t[NLEFT][NTOKEN] = INUM /\ t[NRIGHT][NTOKEN] = INUM)
			t := fold_binary(t, op);
		else ie (t[NLEFT][NTOKEN] = INUM)
			t := fold_left(t, op);
		else if (t[NRIGHT][NTOKEN] = INUM)
			t := fold_right(t, op);
	end
	else ie (op = IDEREF) do
		if (	t[NLEFT] /\ t[NLEFT][NTOKEN] = INUM /\
			t[NLEFT][NVALUE] = 0
		) do
			t[NLEFT] := 0;
			t[NTOKEN] := IGLUE;
		end
	end
	else if (op = IGLUE) do
		if (	t[NRIGHT] /\
			(t[NRIGHT][NTOKEN] = ISAVL \/
			 t[NRIGHT][NTOKEN] = ISAVG) /\
			t[NLEFT] /\ t[NLEFT][NTOKEN] = IGLUE /\
			t[NLEFT][NRIGHT] /\ t[NLEFT][NRIGHT][NTOKEN] = IGLUE /\
			t[NLEFT][NRIGHT][NRIGHT] /\
			\t[NLEFT][NRIGHT][NLEFT] /\
			(t[NLEFT][NRIGHT][NRIGHT][NTOKEN] = IADD \/
			 t[NLEFT][NRIGHT][NRIGHT][NTOKEN] = ISUB))
		do
			t := reduce_inc(t);
		end
	end
	return t;
end


generate(tree) do
	tree := optimize(tree);
	!dumptree(tree, 0);
	emittree(tree);
	cleanup();
end


consop(op, optree, na) do
	var	o;

	if (	\optree \/
		\optree[NLEFT] /\ na = 2
	)
		report("OOPS - too few operands to operation");
	ie (na = 2) do
		o := newnode(op, 0,
			optree[NLEFT][NRIGHT],
			optree[NRIGHT]);
		return newnode(IGLUE, 0, optree[NLEFT][NLEFT], o);
	end
	else do
		optree[NRIGHT] := newnode(op, 0, optree[NRIGHT], 0);
		return optree;
	end
end


moveops(t, p) do
	var	o;

	if (\p[0]) return t;
	o := newnode(IGLUE, 0, t, p[0]);
	p[0] := 0;
	return o;
end


addnode(op, nv, v1, v2, t) do
	var	x;

	t := newnode(IGLUE, 0, t,
		newnode(op, v1, 0, 0));
	x := t[NRIGHT];
	x[NVAL2] := v2;
	x[NNVAL] := nv;
	return t;
end


pass() do
	var	ch[2], v;
	var	lab;
	var	tree, ops;
	var	doinit;

	ch[1] := 0;
	lab := %1;
	tree := 0;
	ops := 0;
	doinit := 0;
	while (1) do
		scan();
		if (	Token \= ISTR /\ Token \= IDATA /\ Token \= IDECL /\
			lab \= %1
		) do
			ie (ops) do
				ops := newnode(IGLUE, 0, ops,
					newnode(ILABEL, lab, 0, 0));
				ops[NRIGHT][NNVAL] := 1;
			end
			else do 
				tree := newnode(ILABEL, lab, 0, tree);
				tree[NNVAL] := 1;
			end
			lab := %1;
		end
		ie (Token = EOF) do
			leave;
		end
		else ie (Token = ILABEL) do
			lab := num();
		end
		else ie (Token = IPOP) do
			tree := moveops(tree, @ops);
			tree := addnode(IPOP, 0, 0, 0, tree);
		end
		else ie (Token = IDECL) do
			mark(lab);
			emit(IDECL);
			emitn(num());
			lab := %1;
		end
		else ie (Token = ILDG) do
			ops := newnode(IGLUE, 0, ops,
				newnode(ILDG, num(), 0, 0));
			ops[NRIGHT][NNVAL] := 1;
		end
		else ie (Token = ILDGV) do
			ops := newnode(IGLUE, 0, ops,
				newnode(ILDGV, num(), 0, 0));
			ops[NRIGHT][NNVAL] := 1;
		end
		else ie (Token = ILDL) do
			ops := newnode(IGLUE, 0, ops,
				newnode(ILDL, num(), 0, 0));
			ops[NRIGHT][NNVAL] := 1;
		end
		else ie (Token = ILDLV) do
			ops := newnode(IGLUE, 0, ops,
				newnode(ILDLV, num(), 0, 0));
			ops[NRIGHT][NNVAL] := 1;
		end
		else ie (Token = IIND) do
			ops := consop(Token, ops, 1);
		end
		else ie (Token = IDEREF) do
			ops := consop(Token, ops, 1);
		end
		else ie (Token = ISAVG) do
			tree := moveops(tree, @ops);
			tree := addnode(ISAVG, 1, num(), 0, tree);
		end
		else ie (Token = ISAVL) do
			tree := moveops(tree, @ops);
			tree := addnode(ISAVL, 1, num(), 0, tree);
		end
		else ie (Token = ISTORE) do
			tree := moveops(tree, @ops);
			tree := addnode(ISTORE, 0, 0, 0, tree);
		end
		else ie (Token = ISTACK) do
			tree := moveops(tree, @ops);
			tree := addnode(ISTACK, 1, num(), 0, tree);
		end
		else ie (Token = IUNSTK) do
			tree := moveops(tree, @ops);
			tree := addnode(IUNSTK, 1, num(), 0, tree);
		end
		else ie (Token = ICALL) do
			v := label();
			semi();
			ops := newnode(IGLUE, 0, ops,
				newnode(ICALL, v, 0, 0));
			ops[NRIGHT][NNVAL] := 2|256;
			ops[NRIGHT][NVAL2] := num();
		end
		else ie (Token = INUM) do
			ops := newnode(IGLUE, 0, ops,
				newnode(INUM, num(), 0, 0));
			ops[NRIGHT][NNVAL] := 1;
		end
		else ie (Token = ISTR) do
			mark(lab);
			string();
			lab := %1;
		end
		else ie (Token = ILDLAB) do
			ops := newnode(IGLUE, 0, ops,
				newnode(ILDLAB, label(), 0, 0));
			ops[NRIGHT][NNVAL] := 1|256;
		end
		else ie (Token = IDATA) do
			if (lab \= %1) mark(lab);
			scan();
			emit(IDATA);
			ie (Token = ILABEL)
				mark(num());
			else
				emitn(Value);
			lab := %1;
		end
		else ie (Token = INEG) do
			ops := consop(Token, ops, 1);
		end
		else ie (Token = ILNOT) do
			ops := consop(Token, ops, 1);
		end
		else ie (Token = IBNOT) do
			ops := consop(Token, ops, 1);
		end
		else ie (Token = IMUL) do
			ops := consop(Token, ops, 2);
		end
		else ie (Token = IDIV) do
			ops := consop(Token, ops, 2);
		end
		else ie (Token = IMOD) do
			ops := consop(Token, ops, 2);
		end
		else ie (Token = IADD) do
			ops := consop(Token, ops, 2);
		end
		else ie (Token = ISUB) do
			ops := consop(Token, ops, 2);
		end
		else ie (Token = IBAND) do
			ops := consop(Token, ops, 2);
		end
		else ie (Token = IBOR) do
			ops := consop(Token, ops, 2);
		end
		else ie (Token = IBXOR) do
			ops := consop(Token, ops, 2);
		end
		else ie (Token = IBSHL) do
			ops := consop(Token, ops, 2);
		end
		else ie (Token = IBSHR) do
			ops := consop(Token, ops, 2);
		end
		else ie (Token = IEQU) do
			ops := consop(Token, ops, 2);
		end
		else ie (Token = INEQU) do
			ops := consop(Token, ops, 2);
		end
		else ie (Token = ILESS) do
			ops := consop(Token, ops, 2);
		end
		else ie (Token = IGRTR) do
			ops := consop(Token, ops, 2);
		end
		else ie (Token = ILTEQ) do
			ops := consop(Token, ops, 2);
		end
		else ie (Token = IGTEQ) do
			ops := consop(Token, ops, 2);
		end
		else ie (Token = IINCG) do
			tree := moveops(tree, @ops);
			v := num();
			semi();
			tree := addnode(IINCG, 2, v, num(), tree);
		end
		else ie (Token = IINCL) do
			tree := moveops(tree, @ops);
			v := num();
			semi();
			tree := addnode(IINCL, 2, v, num(), tree);
		end
		else ie (Token = IBRF) do
			ie (ops)
				ops := addnode(IBRF, 0, 0, 0, ops);
			else
				tree := addnode(IBRF, 0, 0, 0, tree);
		end
		else ie (Token = IBRT) do
			ie (ops)
				ops := addnode(IBRT, 0, 0, 0, ops);
			else
				tree := addnode(IBRT, 0, 0, 0, tree);
		end
		else ie (Token = IJUMP) do
			ie (ops)
				ops := addnode(IJUMP, 1|256, label(), 0, ops);
			else
				tree := addnode(IJUMP, 1|256, label(), 0, tree);
		end
		else ie (Token = IUFOR) do
			tree := moveops(tree, @ops);
			tree := addnode(IUFOR, 0, 0, 0, tree);
		end
		else ie (Token = IDFOR) do
			tree := moveops(tree, @ops);
			tree := addnode(IDFOR, 0, 0, 0, tree);
		end
		else ie (Token = IHDR) do
			tree := moveops(tree, @ops);
			if (tree) generate(tree);
			tree := addnode(IHDR, 0, 0, 0, 0);
		end
		else ie (Token = IEND) do
			tree := moveops(tree, @ops);
			tree := addnode(IEND, 0, 0, 0, tree);
			generate(tree);
			tree := 0;
		end
		else ie (Token = IHALT) do
			tree := moveops(tree, @ops);
			tree := addnode(IHALT, 0, 0, 0, tree);
		end
		else ie (Token = IPUSH) do
			tree := moveops(tree, @ops);
			tree := addnode(IPUSH, 0, 0, 0, tree);
		end
		else ie (Token = INOP) do
			ie (ops)
				ops := addnode(INOP, 0, 0, 0, ops);
			else
				tree := addnode(INOP, 0, 0, 0, tree);
		end
		else ie (Token = IINIT) do
			emit(IINIT);
			doinit := 2;
		end
		else ie (Token = NUMBER) do
			ie (doinit)
				report("input file version is past T3");
			else
				report("unexpected free number");
		end
		else do
			report("unknown symbol");
		end
		if (doinit) doinit := doinit - 1;
	end
	moveops(tree, @ops);
	generate(tree);
end


init() do
	Bp := 0;
	Ep := 0;
	Pos := 0;
	Obp := 0;
	Token := 0;
	Value := 0;
	Nt := 0;
	Maxnodes := 0;
end


do
	init();
	pass();
	bwrite(0);
end

